@@ -1,4 +1,5 @@
 use strict;
+use warnings;
 
 package Test::Deep::Cmp;
 use Carp qw( confess );
@@ -80,4 +81,9 @@ sub string
 	return overload::StrVal($self);
 }
 
+sub reset_arrow
+{
+	return 1;
+}
+
 1;
